/*
 *	Ioctl handler
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_ioctl.c,v 1.4 2000/11/08 05:16:40 davem Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/inetdevice.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp.h"
#include "br_rstp_stm.h"

static int br_ioctl_device(struct net_bridge *br,
			   unsigned int cmd,
			   unsigned long arg0,
			   unsigned long arg1,
			   unsigned long arg2)
{
	if (br == NULL)
		return -EINVAL;

	switch (cmd)
	{
	case BRCTL_ADD_IF:
	case BRCTL_DEL_IF:
	{
		struct net_device *dev;
		int ret;

		dev = dev_get_by_index(arg0);
		if (dev == NULL)
			return -EINVAL;

		if (cmd == BRCTL_ADD_IF)
			ret = br_add_if(br, dev);
		else
			ret = br_del_if(br, dev);

		dev_put(dev);
		return ret;
	}

	case BRCTL_GET_BRIDGE_INFO:
	{
		struct __bridge_info b;
		struct net_bridge_port *p;

		memset(&b, 0, sizeof(struct __bridge_info));
		memcpy(&b.designated_root, &br->rootPriority.RootBridgeID, 8);
		memcpy(&b.bridge_id, &br->BridgeIdentifier, 8);
		b.root_path_cost = br->rootPriority.RootPathCost;
		b.forward_delay = br->rootTimes.FwdDelay;
		b.hello_time = br->rootTimes.HelloTime;
		b.max_age = br->rootTimes.MaxAge;
		b.bridge_forward_delay = br->BridgeTimes.FwdDelay;
		b.bridge_hello_time = br->BridgeTimes.HelloTime;
		b.bridge_max_age = br->BridgeTimes.MaxAge;

		b.topology_change = 0;
		p = br->port_list;
		while (p != NULL) {
			if (br_timer_is_running(&p->tcWhile)) {
				b.topology_change = 1;
				break;
			}
			p = p->next;
		}
		b.topology_change_detected = b.topology_change;

		p = br_get_port_by_portId(br, br->rootPortId);
		b.root_port = (p != NULL ? p->port_no : 0);

		b.stp_enabled = 1;
		b.ageing_time = br->AgeingTime;
		b.gc_interval = 0;                  /* not supported */
		b.hello_timer_value = 0;            /* obsolete: now per-port */
		b.tcn_timer_value = 0;              /* obsolete: now per-port */
		b.topology_change_timer_value = 0;  /* obsolete: now per-port */
		b.gc_timer_value = 0;               /* not supported */

		if (copy_to_user((void *)arg0, &b, sizeof(b)))
			return -EFAULT;

		return 0;
	}

	case BRCTL_GET_PORT_LIST:
	{
		int i;
		int indices[256];

		for (i=0;i<256;i++)
			indices[i] = 0;

		br_get_port_ifindices(br, indices);
		if (copy_to_user((void *)arg0, indices, 256*sizeof(int)))
			return -EFAULT;

		return 0;
	}

	case BRCTL_SET_BRIDGE_FORWARD_DELAY:
		/* 17.28.2 (IEEE 802.1w) */
		if (arg0 >= 4*HZ && arg0 <= 30*HZ) {
			struct net_bridge_port *p;

			br->BridgeTimes.FwdDelay = arg0;

			p = br->port_list;
			while (p != NULL) {
				p->reselect = TRUE;
				p = p->next;
			}

			return 0;
		}
		return -EINVAL;

	case BRCTL_SET_BRIDGE_HELLO_TIME:
		/* 17.28.2 (IEEE 802.1w) */
		if (arg0 >= 1*HZ && arg0 <= 10*HZ) {
			struct net_bridge_port *p;

			br->BridgeTimes.HelloTime = arg0;

			p = br->port_list;
			while (p != NULL) {
				p->reselect = TRUE;
				p = p->next;
			}

			return 0;
		}
		return -EINVAL;

	case BRCTL_SET_BRIDGE_MAX_AGE:
		/* 17.28.2 (IEEE 802.1w) */
		if (arg0 >= 6*HZ && arg0 <= 40*HZ) {
			struct net_bridge_port *p;

			br->BridgeTimes.MaxAge = arg0;

			p = br->port_list;
			while (p != NULL) {
				p->reselect = TRUE;
				p = p->next;
			}

			return 0;
		}
		return -EINVAL;

	case BRCTL_SET_AGEING_TIME:
		/* 7.9.2 (IEEE 802.1d) */
		if (arg0 >= 10*HZ && arg0 <= 1000000*HZ) {
			struct net_bridge_port *p;

			br->AgeingTime = arg0;

			p = br->port_list;
			while (p != NULL) {
				p->reselect = TRUE;
				p = p->next;
			}

			return 0;
		}
		return -EINVAL;

	case BRCTL_SET_GC_INTERVAL:
		return -EINVAL;

	case BRCTL_GET_PORT_INFO:
	{
		struct __port_info p;
		struct net_bridge_port *pt;

		if ((pt = br_get_port(br, arg1)) == NULL)
			return -EINVAL;

		memset(&p, 0, sizeof(struct __port_info));
		memcpy(&p.designated_root, &pt->portPriority.RootBridgeID, 8);
		memcpy(&p.designated_bridge, &pt->portPriority.DesignatedBridgeID, 8);
		p.port_id = pt->portId;
		p.designated_port = pt->portPriority.DesignatedPortID;
		p.path_cost = pt->PortPathCost;
		p.designated_cost = pt->portPriority.RootPathCost;

		/* 17.5 */
		if (pt->forwarding)
			p.state = BR_STATE_FORWARDING;
		else if (pt->learning)
			p.state = BR_STATE_LEARNING;
		else if (pt->role == BR_ROLE_ROOT || pt->role == BR_ROLE_DESIGNATED)
			p.state = BR_STATE_LISTENING;
		else if (pt->dev->flags & IFF_UP)
			p.state = BR_STATE_BLOCKING;
		else
			p.state = BR_STATE_DISABLED;

		p.top_change_ack = pt->tcAck;
		p.config_pending = pt->newInfo;
		p.message_age_timer_value = br_timer_get_residue(&pt->rcvdInfoWhile);
		p.forward_delay_timer_value = br_timer_get_residue(&pt->fdWhile);
		p.hold_timer_value = 0;  // obsolete

		if (copy_to_user((void *)arg0, &p, sizeof(p)))
			return -EFAULT;

		return 0;
	}

	case BRCTL_SET_BRIDGE_STP_STATE:
		return arg0 ? 0 : -EINVAL;

	case BRCTL_SET_BRIDGE_PRIORITY:
		br_stp_set_bridge_priority(br, arg0);
		return 0;

	case BRCTL_SET_PORT_PRIORITY:
	{
		struct net_bridge_port *p;

		if ((p = br_get_port(br, arg0)) == NULL)
			return -EINVAL;
		br_stp_set_port_priority(p, arg1);
		return 0;
	}

	case BRCTL_SET_PATH_COST:
	{
		struct net_bridge_port *p;

		if ((p = br_get_port(br, arg0)) == NULL)
			return -EINVAL;
		br_stp_set_path_cost(p, arg1);
		return 0;
	}

	case BRCTL_GET_FDB_ENTRIES:
		return br_fdb_get_entries(br, (void *)arg0, arg1, arg2);

	case BRCTL_GET_BRIDGE_INFO_RSTP:
	{
		struct __bridge_info_rstp b;
		struct net_bridge_port *p;

		memset(&b, 0, sizeof(struct __bridge_info));
		memcpy(&b.designated_root, &br->rootPriority.RootBridgeID, 8);
		memcpy(&b.bridge_id, &br->BridgeIdentifier, 8);
		b.root_path_cost = br->rootPriority.RootPathCost;
		b.forward_delay = br->rootTimes.FwdDelay;
		b.hello_time = br->rootTimes.HelloTime;
		b.max_age = br->rootTimes.MaxAge;
		b.message_age = br->rootTimes.MessageAge;
		b.bridge_forward_delay = br->BridgeTimes.FwdDelay;
		b.bridge_hello_time = br->BridgeTimes.HelloTime;
		b.bridge_max_age = br->BridgeTimes.MaxAge;

		p = br_get_port_by_portId(br, br->rootPortId);
		b.root_port = (p != NULL ? p->port_no : 0);

		b.force_version = br->ForceVersion;

		b.ageing_time = br->AgeingTime;

		if (copy_to_user((void *)arg0, &b, sizeof(b)))
			return -EFAULT;

		return 0;
	}

	case BRCTL_GET_PORT_INFO_RSTP:
	{
		struct __port_info_rstp p;
		struct net_bridge_port *pt;

		if ((pt = br_get_port(br, arg1)) == NULL)
			return -EINVAL;

		memset(&p, 0, sizeof(struct __port_info));
		memcpy(&p.designated_root, &pt->portPriority.RootBridgeID, 8);
		memcpy(&p.designated_bridge, &pt->portPriority.DesignatedBridgeID, 8);
		p.port_id = pt->portId;
		p.designated_port = pt->portPriority.DesignatedPortID;
		p.path_cost = pt->PortPathCost;
		p.designated_cost = pt->portPriority.RootPathCost;

		p.role = pt->role;
		if (pt->forwarding)
			p.state = BR_RSTP_STATE_FORWARDING;
		else if (pt->learning)
			p.state = BR_RSTP_STATE_LEARNING;
		else
			p.state = BR_RSTP_STATE_DISCARDING;

		p.top_change = pt->tc;
		p.top_change_ack = pt->tcAck;
		p.config_pending = pt->newInfo;
		p.oper_p2pmac = pt->operPointToPointMAC;
		p.oper_edge = pt->operEdge;
		p.sending_rstp = pt->sendRSTP;

		p.message_age_timer_value = br_timer_get_residue(&pt->rcvdInfoWhile);
		p.forward_delay_timer_value = br_timer_get_residue(&pt->fdWhile);
		p.topology_change_timer_value = br_timer_get_residue(&pt->tcWhile);

		if (copy_to_user((void *)arg0, &p, sizeof(p)))
			return -EFAULT;

		return 0;
	}

	case BRCTL_SET_FORCE_VERSION:
		if (arg0 < 0 || arg0 > 2)
			return -EINVAL;
		br->ForceVersion = arg0;
		return 0;

	case BRCTL_SET_P2PMAC:
	{
		struct net_bridge_port *p;

		if ((p = br_get_port(br, arg0)) == NULL)
			return -EINVAL;
		p->operPointToPointMAC = (arg1 ? TRUE : FALSE);
		return 0;
	}

	case BRCTL_SET_EDGE:
	{
		struct net_bridge_port *p;

		if ((p = br_get_port(br, arg0)) == NULL)
			return -EINVAL;
		p->operEdge = (arg1 ? TRUE : FALSE);
		return 0;
	}
	}

	return -EOPNOTSUPP;
}

static int br_ioctl_deviceless(unsigned int cmd,
			       unsigned long arg0,
			       unsigned long arg1)
{
	switch (cmd)
	{
	case BRCTL_GET_VERSION:
		return BRCTL_VERSION;

	case BRCTL_GET_BRIDGES:
	{
		int i;
		int indices[64];

		for (i=0;i<64;i++)
			indices[i] = 0;

		if (arg1 > 64)
			arg1 = 64;
		arg1 = br_get_bridge_ifindices(indices, arg1);
		if (copy_to_user((void *)arg0, indices, arg1*sizeof(int)))
			return -EFAULT;

		return arg1;
	}

	case BRCTL_ADD_BRIDGE:
	case BRCTL_DEL_BRIDGE:
	{
		char buf[IFNAMSIZ];

		if (copy_from_user(buf, (void *)arg0, IFNAMSIZ))
			return -EFAULT;

		buf[IFNAMSIZ-1] = 0;

		if (cmd == BRCTL_ADD_BRIDGE)
			return br_add_bridge(buf);

		return br_del_bridge(buf);
	}
	}

	return -EOPNOTSUPP;
}

DECLARE_MUTEX(ioctl_mutex);

int br_ioctl_deviceless_stub(unsigned long arg)
{
	int err;
	unsigned long i[3];

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	if (copy_from_user(i, (void *)arg, 3*sizeof(unsigned long)))
		return -EFAULT;

	down(&ioctl_mutex);
	err = br_ioctl_deviceless(i[0], i[1], i[2]);
	up(&ioctl_mutex);

	return err;
}

int br_ioctl(struct net_bridge *br, unsigned int cmd, unsigned long arg0, unsigned long arg1, unsigned long arg2)
{
	int err;

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	down(&ioctl_mutex);
	err = br_ioctl_deviceless(cmd, arg0, arg1);
	if (err == -EOPNOTSUPP)
		err = br_ioctl_device(br, cmd, arg0, arg1, arg2);
	up(&ioctl_mutex);

	return err;
}

void br_call_ioctl_atomic(void (*fn)(void))
{
	down(&ioctl_mutex);
	fn();
	up(&ioctl_mutex);
}
