/*
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_private.h,v 1.6.2.1 2001/12/24 00:59:27 davem Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#ifndef _BR_PRIVATE_H
#define _BR_PRIVATE_H

#include <linux/netdevice.h>
#include <linux/miscdevice.h>
#include <linux/if_bridge.h>

#include "br_private_timer.h"

#define BR_HASH_BITS 8
#define BR_HASH_SIZE (1 << BR_HASH_BITS)

typedef struct bridge_id bridge_id;
typedef struct mac_addr mac_addr;
typedef __u16 port_id;

struct bridge_id
{
	unsigned char	prio[2];
	unsigned char	addr[6];
};

struct mac_addr
{
	unsigned char	addr[6];
	unsigned char	pad[2];
};

struct net_bridge_fdb_entry
{
	struct net_bridge_fdb_entry	*next_hash;
	struct net_bridge_fdb_entry	**pprev_hash;
	atomic_t			use_count;
	mac_addr			addr;
	struct net_bridge_port		*dst;
	unsigned long			ageing_timer;
	unsigned			is_local:1;
	unsigned			is_static:1;
};

struct net_bridge_priority_vector
{
	bridge_id RootBridgeID;
	int       RootPathCost;
	bridge_id DesignatedBridgeID;
	port_id   DesignatedPortID;
	port_id   BridgePortID;
};

struct net_bridge_times
{
	unsigned long FwdDelay;
	unsigned long HelloTime;
	unsigned long MaxAge;
	unsigned long MessageAge;
};

typedef enum {
	BR_INFO_DISABLED,
	BR_INFO_RECEIVED,
	BR_INFO_MINE,
	BR_INFO_AGED
} net_bridge_port_info;

typedef enum {
	BR_ROLE_DISABLED = 0,
	BR_ROLE_ALTERNATE,
	BR_ROLE_BACKUP,
	BR_ROLE_DESIGNATED,
	BR_ROLE_ROOT
} net_bridge_port_role;

typedef enum {
	BR_BPDU_TYPE_NONE,
	BR_BPDU_TYPE_CFG,
	BR_BPDU_TYPE_RST,
	BR_BPDU_TYPE_TCN
} net_bridge_bpdu_type;

typedef enum {
	BR_BPDU_ROLE_UNKNOWN,
	BR_BPDU_ROLE_ROOT,
	BR_BPDU_ROLE_ALT_BKP,
	BR_BPDU_ROLE_DESIGNATED
} net_bridge_bpdu_role;

struct net_bridge_bpdu {
	net_bridge_bpdu_type type;

	/* additional contents of config BPDUs... */
	unsigned tc    :1;
	unsigned tcAck :1;

	/* additional contents of RST BPDUs... */
	struct net_bridge_priority_vector priority;
	struct net_bridge_times           times;

	net_bridge_bpdu_role role;

	unsigned proposal   :1;
	unsigned learning   :1;
	unsigned forwarding :1;
	unsigned agreement  :1;
};

typedef enum {
	BR_MSG_SUP_DESIG,
	BR_MSG_REP_DESIG,
	BR_MSG_CONF_ROOT,
	BR_MSG_OTHER
} net_bridge_bpdu_msg_type;

struct net_bridge_port
{
	struct net_bridge_port *next;
	struct net_bridge      *br;
	struct net_device      *dev;
	int                    port_no;
	int                    priority;

	/********************************* RSTP **********************************/

	struct net_bridge_bpdu bpdu;  /* BPDU data, filled in by the input code */

	/* STMs' state representations... */
	int *stm_states;

	/* sub-structs... */
	struct net_bridge_priority_vector designatedPriority;  /* 17.18.2 */
	struct net_bridge_times           designatedTimes;     /* 17.18.3 */
	struct net_bridge_priority_vector msgPriority;         /* 17.18.11 */
	struct net_bridge_times           msgTimes;            /* 17.18.12 */
	struct net_bridge_priority_vector portPriority;        /* 17.18.17 */
	struct net_bridge_times           portTimes;           /* 17.18.18 */

	/* scalar items... */
	net_bridge_port_info     infoIs;        /* 17.18.6 */
	int                      portId;        /* 17.18.16 */
	int                      PortPathCost;  /* 17.16.5 */
	net_bridge_bpdu_msg_type rcvdMsg;       /* 17.18.22 */
	net_bridge_port_role     role;          /* 17.18.30 */
	net_bridge_port_role     selectedRole;  /* 17.18.32 */

#ifdef CONFIG_BRIDGE_CARRIER
	unsigned long upSince;
	struct {
		unsigned long data[CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES];
		unsigned last;
	} carrierChanges;
#endif

	/* flags... */
	unsigned operEdge            :1;  /* 17.18.14 = 18.3.4 (IEEE 802.1t), operEdgePort */
	unsigned operPointToPointMAC :1;  /* 6.4.3 */

	unsigned agreed      :1;  /* 17.18.1 */
	unsigned forward     :1;  /* 17.18.4 */
	unsigned forwarding  :1;  /* 17.18.5 */
	unsigned initPm      :1;  /* 17.18.7 */
	unsigned learn       :1;  /* 17.18.8 */
	unsigned learning    :1;  /* 17.18.9 */
	unsigned mcheck      :1;  /* 17.18.10 */
	unsigned newInfo     :1;  /* 17.18.13 */
	unsigned portEnabled :1;  /* 17.18.15 */
	unsigned proposed    :1;  /* 17.18.19 */
	unsigned proposing   :1;  /* 17.18.20 */
	unsigned rcvdBPDU    :1;  /* 17.18.21 */
	unsigned rcvdRSTP    :1;  /* 17.18.23 */
	unsigned rcvdSTP     :1;  /* 17.18.24 */
	unsigned rcvdTc      :1;  /* 17.18.25 */
	unsigned rcvdTcAck   :1;  /* 17.18.26 */
	unsigned rcvdTcn     :1;  /* 17.18.27 */
	unsigned reRoot      :1;  /* 17.18.28 */
	unsigned reselect    :1;  /* 17.18.29 */
	unsigned selected    :1;  /* 17.18.31 */
	unsigned sendRSTP    :1;  /* 17.18.33 */
	unsigned sync        :1;  /* 17.18.34 */
	unsigned synced      :1;  /* 17.18.35 */
	unsigned tc          :1;  /* 17.18.36 */
	unsigned tcAck       :1;  /* 17.18.37 */
	unsigned tcProp      :1;  /* 17.18.38 */
	unsigned txCount     :1;  /* 17.18.40 */
	unsigned updtInfo    :1;  /* 17.18.41 */

#ifdef CONFIG_BRIDGE_CARRIER
	unsigned carrierLowLevelOK :1;
	unsigned carrierOK         :1;
	unsigned carrierUnsteady   :1;
#endif

	/* timers... */
	struct br_timer fdWhile;        /* 17.15.1 */
	struct br_timer helloWhen;      /* 17.15.2 */
	struct br_timer mdelayWhile;    /* 17.15.3 */
	struct br_timer rbWhile;        /* 17.15.4 */
	struct br_timer rcvdInfoWhile;  /* 17.15.5 */
	struct br_timer rrWhile;        /* 17.15.6 */
	struct br_timer tcWhile;        /* 17.15.7 */
	struct br_timer txWhile;        /* not in spec */

#ifdef CONFIG_BRIDGE_CARRIER
	struct br_timer carrierSteadyTimer;
#endif
};

struct net_bridge
{
	struct net_bridge           *next;
	rwlock_t                    lock;
	struct net_bridge_port      *port_list;
	struct net_device           dev;
	struct net_device_stats     statistics;
	rwlock_t                    hash_lock;
	struct net_bridge_fdb_entry *hash[BR_HASH_SIZE];
	struct timer_list           tick;

	/********************************* RSTP **********************************/

	/* STMs' state representations... */
	int *stm_states;

	/* sub-structs... */
	struct bridge_id                  BridgeIdentifier;  /* 17.17.2 */
	struct net_bridge_priority_vector BridgePriority;    /* 17.17.3 */
	struct net_bridge_times           BridgeTimes;       /* 17.17.4 */
	struct net_bridge_priority_vector rootPriority;      /* 17.17.6 */
	struct net_bridge_times           rootTimes;         /* 17.17.7 */

	/* scalar items... */
	port_id rootPortId;  /* 17.17.5 */

	/* parameters... */
	int ForceVersion;           /* 17.16.1 */
	unsigned long MigrateTime;  /* 17.16.4 */
	int TxHoldCount;            /* 17.16.6 */
	unsigned long AgeingTime;   /* IEEE 802.1d */
};

extern struct notifier_block br_device_notifier;
extern unsigned char bridge_ula[6];

/* br.c */
extern void br_dec_use_count(void);
extern void br_inc_use_count(void);

/* br_device.c */
extern int  br_dev_xmit(struct sk_buff *skb, struct net_device *dev);
extern void br_dev_setup(struct net_device *dev);
extern int br_dev_xmit(struct sk_buff *skb, struct net_device *dev);

/* br_fdb.c */
extern void br_fdb_changeaddr(struct net_bridge_port *p,
		       unsigned char *newaddr);
extern void br_fdb_cleanup(struct net_bridge *br);
extern void br_fdb_delete_by_port(struct net_bridge *br,
			   struct net_bridge_port *p);
extern void br_fdb_switch_ports(struct net_bridge *br,
			   struct net_bridge_port *from,
			   struct net_bridge_port *to);
extern struct net_bridge_fdb_entry *br_fdb_get(struct net_bridge *br,
					unsigned char *addr);
extern void br_fdb_put(struct net_bridge_fdb_entry *ent);
extern int  br_fdb_get_entries(struct net_bridge *br,
			unsigned char *_buf,
			int maxnum,
			int offset);
extern void br_fdb_insert(struct net_bridge *br,
		   struct net_bridge_port *source,
		   unsigned char *addr,
		   int is_local);

/* br_forward.c */
extern void br_deliver(struct net_bridge_port *to,
		struct sk_buff *skb);
extern int  br_dev_queue_push_xmit(struct sk_buff *skb);
extern void br_forward(struct net_bridge_port *to,
		struct sk_buff *skb);
extern int  br_forward_finish(struct sk_buff *skb);
extern void br_flood_deliver(struct net_bridge *br,
		      struct sk_buff *skb,
		      int clone);
extern void br_flood_forward(struct net_bridge *br,
		      struct sk_buff *skb,
		      int clone);

/* br_if.c */
extern int br_add_bridge(char *name);
extern int br_del_bridge(char *name);
extern int br_add_if(struct net_bridge *br,
	      struct net_device *dev);
extern int br_del_if(struct net_bridge *br,
	      struct net_device *dev);
extern int br_get_bridge_ifindices(int *indices,
			    int num);
extern void br_get_port_ifindices(struct net_bridge *br,
			   int *ifindices);

/* br_input.c */
/* This line was manually patched by HT on 29.04.2004 */
extern int br_handle_frame_finish(struct sk_buff *skb);
extern int br_handle_frame(struct sk_buff *skb);

/* br_ioctl.c */
extern void br_call_ioctl_atomic(void (*fn)(void));
extern int br_ioctl(struct net_bridge *br,
	     unsigned int cmd,
	     unsigned long arg0,
	     unsigned long arg1,
	     unsigned long arg2);
extern int br_ioctl_deviceless_stub(unsigned long arg);

/* br_netfilter.c */
extern int br_netfilter_init(void);
extern void br_netfilter_fini(void);

#endif
