/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include "br_rstp_stm.h"

void
br_rstp_stm_init_bridge_states(struct net_bridge *br)
{
	int i;

	for (i = 0; i < BR_RSTP_STMS_BRIDGE; i++)
		br->stm_states[i] = BR_RSTP_STM_STATE_BEGIN;
}

void
br_rstp_stm_init_port_states(struct net_bridge_port *p)
{
	int i;

	for (i = 0; i < BR_RSTP_STMS_PORTS; i++)
		p->stm_states[i] = BR_RSTP_STM_STATE_BEGIN;
}

void
br_rstp_stm_step(struct net_bridge *br)
{
	struct net_bridge_port *p;
	int i;

#ifdef CONFIG_BRIDGE_CARRIER
	p = br->port_list;
	while (p != NULL) {
		int ok = jiffies - p->upSince < 5*HZ ||  // 5s initial tolerance
		         netif_carrier_ok(p->dev);

		 if (p->carrierLowLevelOK != ok) {  // previous vs. current
#if CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES > 0
			p->carrierChanges.last =
				(p->carrierChanges.last+1) %
				    CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES;
			p->carrierChanges.data[p->carrierChanges.last] = jiffies;

			if (!ok && !p->carrierUnsteady &&
			    p->carrierChanges.data[CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES-1] &&
			    p->carrierChanges.data[p->carrierChanges.last] - p->carrierChanges.data[(p->carrierChanges.last + 1) % CONFIG_BRIDGE_CARRIER_EXTENDED_CHANGES] <= CONFIG_BRIDGE_CARRIER_EXTENDED_INTERVAL * HZ)
			{
				p->carrierUnsteady = TRUE;
				printk(KERN_INFO "%s: carrier unsteady on port %i (%s): disabling until steady for %ds\n",
				       p->br->dev.name, p->port_no, p->dev->name,
				       CONFIG_BRIDGE_CARRIER_EXTENDED_REENABLE_DELAY);
			}
#endif

			p->carrierLowLevelOK = ok;
		}

		if (p->carrierOK != ok) {  // propagated vs. current
			if (!ok || br_timer_has_expired(&p->carrierSteadyTimer)) {
				p->carrierOK = ok;

				if (!ok) {
					printk(KERN_INFO
					       "%s: lost carrier on port %i (%s): disabling\n",
					       p->br->dev.name, p->port_no, p->dev->name);
				} else {
					p->carrierUnsteady = FALSE;
					printk(KERN_INFO "%s: carrier steady after loss on port %i (%s): re-enabling\n",
					       p->br->dev.name, p->port_no, p->dev->name);
				}

				p->reselect = TRUE;
			}
		}

		if (!ok) {
			br_timer_set(
			  &p->carrierSteadyTimer,
			  (p->carrierUnsteady
			     ? CONFIG_BRIDGE_CARRIER_EXTENDED_REENABLE_DELAY
			     : CONFIG_BRIDGE_CARRIER_REENABLE_DELAY) * HZ);
		}

		p = p->next;
	}
#endif

	i = 0;
	br_rstp_stm_port_role_selection(br, &br->stm_states[i++]);

	p = br->port_list;
	while (p != NULL) {
		/* 17.13 + 17.27, implem.-specific handling of rate limitation */
		if (br_timer_has_expired(&p->txWhile)) {
			if (p->txCount > 0)
				p->txCount--;
			br_timer_set(&p->txWhile, HZ);  /* 1 second */
		}

		i = 0;
		br_rstp_stm_port_information(p, &p->stm_states[i++]);
		br_rstp_stm_port_protocol_migration(p, &p->stm_states[i++]);
		br_rstp_stm_port_role_transitions(p, &p->stm_states[i++]);
		br_rstp_stm_port_transmit(p, &p->stm_states[i++]);
		br_rstp_stm_port_state_transition(p, &p->stm_states[i++]);
		br_rstp_stm_topology_change(p, &p->stm_states[i++]);

		p = p->next;
	}
}
