/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port Protocol Migration STM of 17.26.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_port_protocol_migration = &step;

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	INIT,
	SEND_RSTP,
	SENDING_RSTP,
	SEND_STP,
	SENDING_STP
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {
		case INIT:
			p->initPm = TRUE;
			p->mcheck = FALSE;
			break;

		case SEND_RSTP:
			if (*state != SEND_RSTP && *state != SENDING_RSTP) {
				printk(KERN_INFO "%s: port %d (%s): transitioning to RSTP mode\n",
					   p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_set(&p->mdelayWhile, p->br->MigrateTime);
			p->mcheck = p->initPm = FALSE;
			p->sendRSTP = TRUE;
			break;

		case SENDING_RSTP:
			p->rcvdRSTP = p->rcvdSTP = FALSE;
			break;

		case SEND_STP:
			if (*state != SEND_STP && *state != SENDING_STP) {
				printk(KERN_INFO "%s: port %d (%s): transitioning to STP mode\n",
					   p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_set(&p->mdelayWhile, p->br->MigrateTime);
			p->sendRSTP = FALSE;
			p->initPm = FALSE;
			break;

		case SENDING_STP:
			p->rcvdRSTP = p->rcvdSTP = FALSE;
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	if (state != BEGIN && !p->portEnabled && !p->initPm)
		enter(INIT);
	else {
		switch (state) {
			case BEGIN:
				enter(INIT);
				break;

			case INIT:
				if (p->portEnabled) {
					if (p->br->ForceVersion >= 2)
						enter(SEND_RSTP);
					else
						enter(SEND_STP);
				}
				break;

			case SEND_RSTP:
				enter(SENDING_RSTP);
				break;

			case SENDING_RSTP:
				if (!br_timer_has_expired(&p->mdelayWhile) &&
				    (p->rcvdSTP || p->rcvdRSTP))
				{
					enter(SENDING_RSTP);
				} else if ((br_timer_has_expired(&p->mdelayWhile) && p->rcvdSTP) ||
				           p->br->ForceVersion < 2)
				{
					enter(SEND_STP);
				} else if (p->mcheck)
					enter(SEND_RSTP);
				break;

			case SEND_STP:
				enter(SENDING_STP);
				break;

			case SENDING_STP:
				if (!br_timer_has_expired(&p->mdelayWhile) && (p->rcvdSTP || p->rcvdRSTP))
					enter(SENDING_STP);
				else if ((br_timer_has_expired(&p->mdelayWhile) && p->rcvdRSTP) ||
				         p->mcheck)
				{
					enter(SEND_RSTP);
				}
				break;

			default:
				printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
					   p->br->dev.name, __FILE__, state);
				enter(BEGIN);
		}
	}
#undef enter
}
